# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionHandler do
  @moduledoc """
  This module is responsible for handling incoming packets and sending outgoing packets. It keeps track of what state the connection is in and which game should
  receive the packets.
  """
  require Logger

  @spec child_spec(:gen_tcp.socket()) :: Supervisor.child_spec()
  def child_spec(socket) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start, [socket, Amethyst.ConnectionState.Handshake, 0]}
    }
  end

  @spec start(:gen_tcp.socket(), atom(), integer()) :: no_return()
  def start(socket, connstate, version) do
    {:ok, spawn(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, connstate, version, %{})
    end)}
  end

  @spec start_link(:gen_tcp.socket(), atom(), integer()) :: no_return()
  def start_link(socket, connstate, version) do
    {:ok, spawn_link(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, connstate, version, %{})
    end)}
  end

  @spec loop(:gen_tcp.socket(), atom(), integer(), map()) :: no_return()
  defp loop(socket, connstate, version, state) do
    receive do
      :closed ->
        Logger.info("Connection #{inspect(socket)} closed.")
        Process.exit(self(), :normal)
      {:disconnect, reason} ->
        disconnect(socket, reason, connstate, version)
        Process.exit(self(), :normal)
      {:set_state, newstate} ->
        Logger.debug("Switching to state #{newstate} from #{connstate}")
        loop(socket, newstate, version, state)
      {:set_version, newversion} ->
        Logger.debug("Switching to version #{newversion} from #{version}")
        loop(socket, connstate, newversion, state)
      {:set_position, position} ->
        Logger.debug("Updating client position to #{inspect(position)}")
        prev_position = Map.get(state, :position)
        state = Map.put(state, :position, position)
        # If there was no prev position, we consider that we
        # definitely moved
        prev_cp = if prev_position == nil do nil else chunk_pos(elem(prev_position, 0), elem(prev_position, 1)) end
        cp = chunk_pos(elem(position, 0), elem(position, 1))
        if prev_cp != cp do
          Logger.debug("Client entered new chunk #{inspect(cp)}")
          # We changed chunk borders, update center chunk and begin sending new chunks
          send(self(), {:send_packet, %{
            packet_type: :set_center_chunk,
            chunk_x: elem(cp, 0),
            chunk_z: elem(cp, 1)
          }})
          # Figure out which new chunks are visible
          prev_chunks =
            if prev_cp == nil do
              MapSet.new([])
            else
              MapSet.new(visible_chunks_from(elem(prev_cp, 0), elem(prev_cp, 1), Map.get(state, :view_distance, 16)))
            end
          chunks = MapSet.new(visible_chunks_from(elem(cp, 0), elem(cp, 1), Map.get(state, :view_distance, 16)))
          new_chunks = MapSet.difference(chunks, prev_chunks)
          # We can process all chunks in parallel
          me = self()
          ts = state |> Map.get(:game) |> Map.get(:refs) |> Map.get(:task_supervisor)
          Task.Supervisor.async(ts, fn ->
            Task.Supervisor.async_stream(ts,
            new_chunks,
            fn chunk -> process_chunk(me, chunk, state) end,
            [ordered: false]
          ) |> Stream.run() end)
        end
        loop(socket, connstate, version, state)
      {:send_packet, packet} ->
        Logger.debug("Sending packet #{inspect(packet)}")
        send_packet(socket, connstate, packet, version)
        loop(socket, connstate, version, state)
      after 0 ->
        receive do
          {:packet, id, data} ->
            state = handle_packet(id, data, connstate, version, state)
            loop(socket, connstate, version, state)
        end
    end
  end

  defp chunk_pos(x, z) do
    {div(floor(x), 16), div(floor(z), 16)}
  end

  defp visible_chunks_from(x, z, view_distance) do
    {cx, cz} = chunk_pos(x, z)
    (cx - view_distance - 3 .. cx + view_distance + 3) |> Enum.flat_map(fn ix ->
      (cz - view_distance - 3 .. cz + view_distance + 3) |> Enum.map(fn iz ->
        {ix, iz}
      end)
    end)
  end

  defp process_chunk(to, chunk, state) do
    import Amethyst.NBT.Write
    alias Amethyst.Minecraft.Write

    chunk_array = Amethyst.API.Game.chunk(Map.get(state, :game), chunk)
    {cx, cz} = chunk

    # TODO: Actually do heightmaps
    # TODO: Doing all this processing could be at home somewhere else
    heightmaps = compound(%{})

    data = Enum.chunk_every(chunk_array, 16, 16, 0) # 0 -> air
        |> Enum.reduce("", fn chunk_section, acc ->
          blocks = chunk_section |> List.flatten()
          block_count = blocks |> Enum.filter(&(&1 != 0)) |> length

          # Put together the palette
          unique_blocks = MapSet.new(blocks)
          min_bpe = MapSet.size(unique_blocks) |> :math.log2() |> ceil()
          paletted_container_data = case min_bpe do
            0 -> <<0::8>> <>
            # SINGLE VALUED
              Write.varint(MapSet.to_list(unique_blocks) |> List.first()) <>
              Write.varint(0) # No data, empty pallette
            min_bpe when min_bpe in 1..8 ->
            # INDIRECT
              # Minimum bpe accepted by minecraft is 4
              bpe = max(min_bpe, 4)
              palette = MapSet.to_list(unique_blocks) |>
                Enum.with_index() |>
                Map.new(fn {i, v} -> {v, i} end)
              paletted_blocks = blocks |>
                Enum.map(&(Map.get(palette, &1)))
              paletted_data = long_aligned_bit_string_reduce(paletted_blocks, bpe)

              Write.ubyte(bpe) <>
              Write.varint(map_size(palette)) <>
              Enum.reduce(0..(map_size(palette)-1), "", fn i, acc ->
                acc <> Write.varint(Map.get(palette, i))
              end) <>
              Write.varint(floor(bit_size(paletted_data) / 64)) <>
              paletted_data
            _ ->
            # DIRECT
              data = long_aligned_bit_string_reduce(blocks, 15)
              Write.ubyte(15) <>
              Write.varint(floor(bit_size(data) / 64)) <>
              data
          end

          # TODO: Send biome data, if that even makes sense
          acc <> Write.short(block_count) <> paletted_container_data <> <<0::8, 0::8>>
        end)

      send(to, {:send_packet, %{
        packet_type: :chunk_data_and_update_light,
        chunk_x: cx, chunk_z: cz,
        heightmaps: heightmaps,
        data: data,
        block_entities: [],
        # TODO: Light
        sky_light_mask: <<0>>,
        block_light_mask: <<0>>,
        empty_sky_light_mask: <<0>>,
        empty_block_light_mask: <<0>>,
        sky_light_arrays: [],
        block_light_arrays: []
      }})
      :ok
  end

  defp long_aligned_bit_string_reduce(values, bpe) do
    values |> Enum.reduce(<<>>, fn value, acc ->
      next = <<acc::bitstring, value::size(bpe)-big>>
      # man i hope they dont suddenly change the size of a long
      if rem(bit_size(next), 64) + bpe < 64 do
        # gotta pad it
        <<next::bitstring, 0::big-size(64 - rem(bit_size(next), 64))>>
      else
        next
      end
    end)
  end

  defp handle_packet(id, data, connstate, version, state) do
    try do
      packet = connstate.deserialize(id, version, data)
      case connstate.handle(packet, version, state) do
        :ok -> state
        {:error, reason} ->
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{reason}")
          send(self(), {:disconnect, "§cError handling packet #{inspect(id, base: :hex)}:\n#{reason}"})
          state
        newstate ->
          if is_map(newstate) do
            newstate
          else
            Logger.warning("State change to #{inspect(newstate)} is not a map! Did you forget to return :ok?")
            state
          end
      end
    rescue
      e ->
        Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        send(self(), {:disconnect, "§cError handling packet #{inspect(id, base: :hex)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"})
        state
    end
  end

  defp send_packet(socket, connstate, packet, version) do
    try do
      data = connstate.serialize(packet, version)
      length = byte_size(data) |> Amethyst.Minecraft.Write.varint()
      :gen_tcp.send(socket, length <> data)
    rescue
      e ->
        Logger.error("Error sending packet #{inspect(packet)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        send(self(), {:disconnect, "§cError sending packet #{inspect(packet)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"})
    end
  end

  defp disconnect(socket, reason, connstate, version) do
    Logger.info("Disconnecting connection #{inspect(socket)}")
    Logger.debug("Disconnecting connection #{inspect(socket)}: #{reason}")
    case connstate.disconnect(reason) do
      nil -> nil
      packet -> send_packet(socket, connstate, packet, version)
    end
    :gen_tcp.close(socket)
  end
end
